// This file is part of the brownmove simulation package
// 
// (C) 2003 - 2009 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove

#ifndef BONDSHAPE_H
#define BONDSHAPE_H

#include <string>
#include <vector>

#include "types.h"
#include "constants.h"
#include "vectormath.h"

#include "shape.h"


namespace brown{
	//! @author tg
	//
	//! This class implements elastic (harmonic) bonds between proteins, so that
	//! flexible (macro)molecules can be constructed. Each bond has an index, which 
	//! has to be the same on the two interaction partners.
	//! The interaction parameters (average distance and spring constant) are stored 
	//! in a vector of structures and indexed via the index of the interaction partners.
	
	class BondShape : public Shape {
		
		private:
		vector<bondPos> Hooks;
		vector<bondPos> Actual_Hooks;
		std::vector<bondParameter> *Params_Vector;
		
		public:
		BondShape();
		BondShape(const BondShape &other);
		~BondShape();
		
		//! restores the (absolute and relative) positions of the parts from fullPosition
		virtual void recalculate(const fullPosition &p);

		//! returns some string for testing
		virtual std::string describe();
		
		//! Adds a new hook for a spring with position and index to the vector of hooks,
		//! also calculates the relative position of the hook w.r.t the molecule center.
		void addHook(bondPos bp, const position &moleculeCenter);
		
		//! Set the pointer to the vector of bond-parameter-structs
		void setParameterVector(std::vector<bondParameter> *theVector) {
			Params_Vector = theVector;
		}
		
		//! Calculate the forces and torques from the relative positions of this and the other BondShape.
		bool calculateImpact(fullForce &myForce, fullForce &otherForce, BondShape *bs);
		
		//! Returns the number of hooks.
		int size();
		
		//! Dumps the current state of the bonds in browndef format
		void dumpStructure();
	};
}

#endif /* BONDSHAPE */
