// This file is part of the brownmove simulation package
// 
// (C) 2003 - 2009 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove

#include "gestalt.h"


namespace brown{
	
	
Gestalt::Gestalt(std::string lbl)
{
	myLabel = lbl;
	
	myEstatShape = NULL;
	myGeomShape  = NULL;
	myvdWShape = NULL;
	mybondShape = NULL;
	myHiShape = NULL;
	myExternalShape = NULL;
	
	myWallsPosition = NULL;
	myWallsNormal = NULL;
	
	resetTotalForce();
	immobilizeCounter = 0;
}

//! copies from template and keeps pointer to that template.
//! so use with care and copy only from Prototypes!
Gestalt::Gestalt(const Gestalt &other)
{
	myLabel = other.myLabel;
	
	myWallsPosition = other.myWallsPosition;
	myWallsNormal = other.myWallsNormal;
	
	myPosition = other.myPosition;
	myActualPosition = other.myActualPosition;
	
	myEstatShape = NULL;
	myGeomShape  = NULL;
	myvdWShape = NULL;
	mybondShape = NULL;
	myHiShape = NULL;
	myExternalShape = NULL;
	
	if (other.getGeomShape()) {
		myGeomShape = new GeomShape(*other.getGeomShape());
	}
	if (other.getEstatShape()) {
		myEstatShape = new EstatShape(*other.getEstatShape());
	}
	if (other.getvdWShape()) {
		myvdWShape = new vdWShape(*other.getvdWShape());
	}
	if (other.getBondShape()) {
		mybondShape = new BondShape(*other.getBondShape());
	}
	if (other.getHiShape()) {
		myHiShape = new HiShape(*other.getHiShape());
	}
	if (other.getExternalShape()) {
		myExternalShape = new ExternalShape(*other.getExternalShape());
	}
	
	resetTotalForce();
	immobilizeCounter = other.immobilizeCounter;
}

Gestalt::~Gestalt(){
	// Check for Shapes and tear 'em down
	if (myGeomShape) {
		delete myGeomShape; 
	}
	if (myEstatShape) {
		delete myEstatShape;
	}
	if (myvdWShape) {
		delete myvdWShape;
	}
	if (mybondShape) {
		delete mybondShape;
	}
	if (myHiShape) {
		delete myHiShape;
	}
	if (myExternalShape) {
		delete myExternalShape;
	}
}



std::string Gestalt::describe() {
	ostringstream ausgabe("", ios::out);
	int i=0;
	
	ausgabe << "gestalt >>" << myLabel <<"<<, positions:";
	ausgabe << " : "<<myActualPosition.pos[0]<<" "<<myActualPosition.pos[1]<<" "<<myActualPosition.pos[2] ; 
	ausgabe << " / "<<myActualPosition.rot.phi<<" "<<myActualPosition.rot.n[0]<<" "<<myActualPosition.rot.n[1]<<" "<<myActualPosition.rot.n[2] ; 
	ausgabe <<" :";
	ausgabe << " : "<<myPosition.pos[0]<<" "<<myPosition.pos[1]<<" "<<myPosition.pos[2] ; 
	ausgabe << " / "<<myPosition.rot.phi<<" "<<myPosition.rot.n[0]<<" "<<myPosition.rot.n[1]<<" "<<myPosition.rot.n[2] ; 

	return(ausgabe.str());
}


void Gestalt::dumpStructure()
{
	std::cout << "\n\tGestalt " << myLabel << std::endl;
	std::cout << "\t\tposition "<<myActualPosition.pos[0]<<" "<<myActualPosition.pos[1]<<" "<<myActualPosition.pos[2]<<"   "<<myActualPosition.rot.phi<<" "<<myActualPosition.rot.n[0]<<" "<<myActualPosition.rot.n[1]<<" "<<myActualPosition.rot.n[2]<<std::endl;
	std::cout << "\t\tD0 " << myGeomShape->getD0Trans() <<" "<< myGeomShape->getD0Rot() <<" "<< myGeomShape->getMass() <<" "<< myGeomShape->getRadius() << std::endl;
	
	// Internals
	if(mybondShape) {
		mybondShape->dumpStructure();
	}
	if(myEstatShape) {
		myEstatShape->dumpStructure();
	}
	if(myvdWShape) {
		myvdWShape->dumpStructure();
	}
	if(myHiShape) {
		myHiShape->dumpStructure();
	}
	
	std::cout << "\tEnd" << std::endl;
}



void Gestalt::positionToSnapshot(FILE *fp, const char *eiDi, const long int id_number)
{
	fprintf(fp, "%s\t%ld\t%s\t%.5e\t%.5e\t%.5e\t%.5e\t%.5e\t%.5e\t%.5e\n", 
			eiDi, id_number, myLabel.c_str(),
			myActualPosition.pos[0], myActualPosition.pos[1], myActualPosition.pos[2],  
			myActualPosition.rot.phi, myActualPosition.rot.n[0], myActualPosition.rot.n[1], myActualPosition.rot.n[2]);
}



void Gestalt::resetTotalForce() 
{
	doublevec_set(totalForce.f,            0.0);
	doublevec_set(totalForce.t,            0.0);
	doublevec_set(randomForce.f,           0.0);
	doublevec_set(randomForce.t,           0.0);
	doublevec_set(externalForce.f,         0.0);
	doublevec_set(externalForce.t,         0.0);
	doublevec_set(ext_Correction_Force.f,  0.0);
	doublevec_set(ext_Correction_Force.t,  0.0);
	doublevec_set(rand_Correction_Force.f, 0.0);
	doublevec_set(rand_Correction_Force.t, 0.0);
	doublevec_set(correction_Factor_Trans, 0.0);
	doublevec_set(correction_Factor_Rot,   0.0);
}


void Gestalt::setPositionNormal(position *p, direction *n) 
{
	myWallsPosition = p;
	myWallsNormal = n;
	
	if(myGeomShape) {
		myGeomShape->setPositionNormal(p, n);
	}
	if(myEstatShape) {
		myEstatShape->setPositionNormal(p, n);
	}
	if(myvdWShape) {
		myvdWShape->setPositionNormal(p, n);
	}
}



void Gestalt::teleport(const fullPosition &fp)
{
	quaternion myQ, dQ, ergQ;
	
	// rotation part
	rotation2quaternion(myQ, myPosition.rot);
	rotation2quaternion(dQ, fp.rot);
	quaternion_mult(ergQ, dQ, myQ);
	quaternion2rotation(myActualPosition.rot, ergQ);
	
	// translation part
	doublevec_set(myActualPosition.pos, myPosition.pos);
	doublevec_rotate(myActualPosition.pos, fp.rot);
	doublevec_add(myActualPosition.pos, fp.pos);
	
	recalculate(myActualPosition);
}


// wrapper calling move(dx, dphi) below
void Gestalt::move(const fullPosition &dfp) {
	move(dfp.pos, dfp.rot);
}


void Gestalt::move(const position &dx, const rotation &dphi)
{
	quaternion myQ, dQ, ergQ;
	
	// displacement part: shift positon
	doublevec_add(myActualPosition.pos, dx);

	// rotation part: rotate in place by dphi
	rotation2quaternion(myQ, myActualPosition.rot);
	rotation2quaternion(dQ, dphi);
	quaternion_mult(ergQ, dQ, myQ);
	quaternion2rotation(myActualPosition.rot, ergQ);
	
	// propagate the new position to internals
	recalculate(myActualPosition);
}

	

void Gestalt::recalculate(const fullPosition &fp) 
{
	// just pass it on
	if(myGeomShape) {
		myGeomShape->recalculate(fp);
	}
	if(myEstatShape) {
		myEstatShape->recalculate(fp);
	}
	if(myvdWShape) {
		myvdWShape->recalculate(fp);
	}
	if(mybondShape) {
		mybondShape->recalculate(fp);
	}
	if(myHiShape) {
		myHiShape->recalculate(fp);
	}
	if(myExternalShape) {
		myExternalShape->recalculate(fp);
	}
}


bool Gestalt::isColliding(const Gestalt *other) {
	return(myGeomShape->checkCollision(other->getGeomShape()));
}


void Gestalt::talkToGestalt(Gestalt &otherGest, bool checkBonds, const bool isPeriodic[3], const doublevec &cellSize) {
	otherGest.calculateImpact(externalForce, this, checkBonds, isPeriodic, cellSize);
}

	

void Gestalt::calculateImpact(fullForce &otherForce, Gestalt *g, bool checkBonds, const bool isPeriodic[3], const doublevec &cellSize)
{
	// std::cerr << "Gestalt::calculateImpact: Gestalt " << g->describe() << " vs. " << describe() << std::endl;
	
	bool is_bonded = false;
	
	// only check for bonds if both gestalten live in the same molecule, this decision was taken further up
	if(checkBonds) {
		if(mybondShape) {
			BondShape *otherBS;
			if((otherBS = g->getBondShape())) {
				is_bonded = mybondShape->calculateImpact(externalForce, otherForce, otherBS);
			}
		}
	}
	
	if(myEstatShape) {
		EstatShape *otherES;
		if((otherES = g->getEstatShape())) {
			myEstatShape->calculateImpact(externalForce, otherForce, otherES, isPeriodic, cellSize);
		}
	}
	if(myvdWShape) {
		vdWShape *otherVS;
		if((otherVS = g->getvdWShape())) {
			myvdWShape->calculateImpact(externalForce, otherForce, otherVS, isPeriodic, cellSize);
		}
	}
}


void Gestalt::initRandomForces(void) {
	myGeomShape->initRandomForces(randomForce);
}


long int Gestalt::initHiForces() {
	if(myHiShape) {
		myHiShape->initForces(randomForce, externalForce);
		return(1L);
	}
	return(0L);
}


double Gestalt::talkHi(Gestalt &otherGest, const bool isPeriodic[3], const doublevec &cellSize) {
	if(myHiShape) {
		HiShape *otherHiS;
		if((otherHiS = otherGest.getHiShape())) {
			bool includeRotation = (myGeomShape->getDoesRotate() || otherGest.getGeomShape()->getDoesRotate());
			
			#ifndef HI_WITH_ROTATION
			includeRotation = false;		// ### HI-Change ###
			#endif
			
			return(myHiShape->getHiCorrections(otherHiS, includeRotation));
		}
	}
	return(-1.0);	// tell that there is no HiShape
}


void Gestalt::correctDamping() {
	myGeomShape->correctDamping(externalForce, randomForce);
}



void Gestalt::doEffectiveForces(const double &beta_prime) 
{
	if(myHiShape && (beta_prime > 0.0)) {						// beta > 0 only for more than 1 HiShapes
		myHiShape->doEffectiveForces(totalForce, beta_prime);
	}
	else	// just add up random and external forces
	{
		doublevec_set(totalForce.f, randomForce.f);
		doublevec_add(totalForce.f, externalForce.f);
		doublevec_set(totalForce.t, randomForce.t);
		doublevec_add(totalForce.t, externalForce.t);
	}
}



void Gestalt::makeBrownianStep()
{
	fullPosition dfp;
	
	myGeomShape->getDisplacement(dfp, totalForce.f, totalForce.t);	// how far to go with the current forces?
	move(dfp.pos, dfp.rot);											// go!
	resetTotalForce();												// forces have been used => reset forces before the next step
}



void Gestalt::calculateExternalForce(fullForce &myForce) {
	if(myExternalShape) {
		myExternalShape->calculateImpact(myForce);
	}
}

}
