// This file is part of the brownmove simulation package
// 
// (C) 2003 - 2009 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove

#include "types.h"
#include "vectormath.h"

namespace brown {
	
	bool fullPosition::operator==(const fullPosition &other){
		return(doublevec_equals(pos, other.pos) && doublevec_equals(rot.n, other.rot.n) && rot.phi==other.rot.phi);
	}


	void fullForce::operator+=(const fullForce &theDelta) {
		doublevec_add(f, theDelta.f);
		doublevec_add(t, theDelta.t);
	}
	
	chargePos::chargePos()
	:ch(0.),r(0.) {
		doublevec_set (pos, 0.0);
	}  
	
	chargePos::chargePos(double p1, double p2, double p3, charge thec, double ther)
	:ch(thec),r(ther) {
		doublevec_set(pos, p1, p2, p3);
		doublevec_set(dr, 0.0);
	}


	vdWPos::vdWPos()  {
		r = 0.0;
		color = 0L;
		doublevec_set(pos, 0.0);
		doublevec_set(dr, 0.0);
	}

	vdWPos::vdWPos(double p1, double p2, double p3, double ther, long int thec) 
	{
		r = ther;
		color = thec;
		doublevec_set(pos, p1, p2, p3);
		doublevec_set(dr, 0.0);
	}

	
	vdWBlock::vdWBlock() {
		doublevec_set(pos, 0.0);
		doublevec_set(ex, 1.0, 0.0, 0.0);
		doublevec_set(ey, 0.0, 1.0, 0.0);
		doublevec_set(ez, 0.0, 0.0, 1.0);
		lx = ly = lz = 0.0;
		r = 0.0;
		color = -1;
	}
	
	
	vdWBlock::vdWBlock(const position &_p, const doublevec &_ex, const doublevec &_ey, const doublevec &_ez,
					 const double &_lx, const double &_ly, const double &_lz, const double &_r) {
		doublevec_set(pos, _p);
		doublevec_set(ex, _ex);
		doublevec_set(ey, _ey);
		doublevec_set(ez, _ez);
		lx = _lx; ly = _ly; lz = _lz;
		r = _r;
	}

	
	bool vdWBlock::getDistance(doublevec &normal, double &distance, const position &thePos, const double &theR)
	{
		doublevec distV;
		
		doublevec_set(distV, thePos);
		doublevec_sub(distV, pos);
		
		// distances from the center in the three directions
		double dx = doublevec_scalprod(distV, ex);
		double dy = doublevec_scalprod(distV, ey);
		double dz = doublevec_scalprod(distV, ez);
		double dr = sqrt(dy*dy + dz*dz);
		
		// now check...
		// above or below plane?
		double hoehe_x = fabs(dx) - lx;
		   
		// inside the cylinder?
		if(dr <= r) {
			doublevec help;
			
			doublevec_set(help, ey);
			doublevec_mult(help, -dy);
			doublevec_set(normal, ez);
			doublevec_mult(normal, -dz);
			doublevec_add(normal, help);
			
			double inside_dist = r - dr;			// distance between sphere center and cylinder wall
			
			if(hoehe_x <= 0.0) {					// center of sphere inside the cylinder, we're done
				if(dr > 1e-6) {
					doublevec_mult(normal, 1.0/dr);
				} else {
					doublevec_set(normal, 1.0, 0.0, 0.0);	// for such a short vector the direction is irrelevant
				}
				distance = inside_dist - theR;		// open space between sphere and cylinder wall
				return(true);
			}
			
			// above the block...
			double above_r = sqrt(hoehe_x*hoehe_x + inside_dist*inside_dist);
			
			doublevec_mult(normal, inside_dist);	// rescale radial component
			
			doublevec_set(help, ex);
			if(dx < 0.0) {							// sphere is below block
				doublevec_mult(help, -1.0);
			}
			doublevec_mult(help, hoehe_x);			// rescaled component off the face of the block
			
			doublevec_add(normal, help);
			if(above_r > 1e-6) {
				doublevec_mult(normal, 1.0 / above_r);
			} else {
				doublevec_set(normal, 1.0, 0.0, 0.0);
			}
			distance = above_r - theR;
			
			return(true);							// finally done... uff
		}
		
		// check if above or below block within +-ly, +-lz
		else if(fabs(dy) <= ly && fabs(dz) <= lz)
		{
			doublevec_set(normal, ex);
			if(dx < 0.0) {
				doublevec_mult(normal, -1.0);
			}
			distance = fabs(dx) - lx - theR;
			
			return(true);
		}
		
		// outside => do nothing
		return(false);
	}
	
	
	geomPos::geomPos(double p1, double p2, double p3, double ther, long int thec)
	: r(ther), color(thec) {
		doublevec_set(pos,p1,p2,p3);
		doublevec_set(dr,0.0);
	}

	
	geomPos::geomPos() : r(0.0), color(0L) {
		doublevec_set(pos,0.0);
		doublevec_set(dr, 0.0);
	}

	
	bondPos::bondPos() : idx(0L) {
	    doublevec_set(pos, 0.0);
	    doublevec_set(dr, 0.0);
	}
	
	
	externalPos::externalPos() {
		doublevec_set(pos, 0.0);
		doublevec_set(dr, 0.0);
		interaction = NULL;
	}
}

